/*
**
** Playing with an LCD display
*/

#include <p24Fxxxx.h>

#if defined( __PIC24FJ256GB110__ ) || defined( __PIC24FJ256GB106__ )
#define PLL_96MHZ_ON    0xF7FF
_CONFIG1(ICS_PGx2 & JTAGEN_OFF & FWDTEN_OFF)        // JTAG off, watchdog timer off
_CONFIG2( FNOSC_PRIPLL & PLL_96MHZ_ON & PLLDIV_DIV2 & OSCIOFNC_OFF & FCKSM_CSDCMD & POSCMOD_XT)
#else                   // PIC24FJ128GA010 and PIC24FJ256GA110
_CONFIG1( ICS_PGx2 & JTAGEN_OFF & GCP_OFF & GWRP_OFF & COE_OFF & FWDTEN_OFF) 
_CONFIG2( FNOSC_PRIPLL & OSCIOFNC_OFF & FCKSM_CSDCMD & POSCMOD_HS)
#endif

#define LCDDATA 1
#define LCDCMD  0
#define PMDATA  PMDIN1

void initLCD( void)
{
    // PMP initialization 
    PMCON = 0x83BF;             // Enable the PMP, long waits
    PMMODE = 0x3FF;             // Master Mode 1
	PMAEN = 0x0001;             // PMA0 enabled
    
    // init TMR1
    T1CON = 0x8030;             // Fosc/2, prescaled 1:256, 16us/tick

    // wait for >30ms
    TMR1 = 0; while( TMR1<2000);// 2000 x 16us = 32ms   
    
    //initiate the HD44780 display 8-bit init sequence
    PMADDR = LCDCMD;            // command register
    PMDATA = 0b00111000;        // 8-bit interface, 2 lines, 5x7
    TMR1 = 0; while( TMR1<3);   // 3 x 16us = 48us   
    
    PMDATA = 0b00001100;        // display ON, cursor off, blink off
    TMR1 = 0; while( TMR1<3);   // 3 x 16us = 48us   
    
    PMDATA = 0b00000001;        // clear display
    TMR1 = 0; while( TMR1<100); // 100 x 16us = 1.6ms   
    
    PMDATA = 0b00000110;        // increment cursor, no shift
    TMR1 = 0; while( TMR1<100); // 100 x 16us = 1.6ms   
} // initLCD


char readLCD( int addr)
{
    int dummy;
    while( PMMODEbits.BUSY);    // wait for PMP to be available
    PMADDR = addr;              // select the command address
    dummy = PMDATA;            // initiate a read cycle, dummy read
    while( PMMODEbits.BUSY);    // wait for PMP to be available
    return( PMDATA);            // read the status register
    
} // readLCD

#define busyLCD() readLCD( LCDCMD) & 0x80
#define addrLCD() readLCD( LCDCMD) & 0x7F
#define getLCD()  readLCD( LCDDATA)


void writeLCD( int addr, char c)    
{
    while( busyLCD());
    while( PMMODEbits.BUSY);    // wait for PMP to be available
    PMADDR = addr;
    PMDATA = c;
} // writeLCD
    

#define putLCD( d)  writeLCD( LCDDATA, (d))
#define cmdLCD( c)  writeLCD( LCDCMD, (c))

#define homeLCD()   writeLCD( LCDCMD, 2)    
#define clrLCD()    writeLCD( LCDCMD, 1)
#define setLCDG( a) writeLCD( LCDCMD, (a & 0x3F) | 0x40)
#define setLCDC( a) writeLCD( LCDCMD, (a & 0x7F) | 0x80)

void putsLCD( char *s)
{
    while( *s) putLCD( *s++);
} //putsLCD



#define TFLY 9000       // 9000 x 16us = 144ms
#define DELAY() TMR1=0; while( TMR1<TFLY)


main( void)
{
    int i;
        
    // initializations
    initLCD();
    
    // put a title on the first line    
    putsLCD( "Flying the PIC24");

    // generate two new characters
    setLCDG(0);
    putLCD( 0b00010);
    putLCD( 0b00010);
    putLCD( 0b00110);
    putLCD( 0b11111);
    putLCD( 0b00110);
    putLCD( 0b00010);
    putLCD( 0b00010);
    putLCD( 0);     // alignment
    
    putLCD( 0b00000);
    putLCD( 0b00100);
    putLCD( 0b01100);
    putLCD( 0b11100);
    putLCD( 0b00000);
    putLCD( 0b00000);
    putLCD( 0b00000);
    putLCD( 0);     // alignment
    
    // main loop
    while( 1)
    {
        // the entire plane appears at the right margin
        setLCDC(0x40+14);
        putLCD( 0); putLCD( 1);
        DELAY();
        
        // fly fly fly (right to left)
        for( i=13; i>=0; i--)
        {
            setLCDC(0x40+i);      // set the cursor to the next position
            putLCD(0); putLCD(1); // new airplane
            putLCD(' ');		 // erase the previous tail
            DELAY();
        }
        
        // the tip disappears off the left margin, only the tail is visible
        setLCDC(0x40);
        putLCD( 1); putLCD(' ');
        DELAY();       

        // erase the tail
        setLCDC(0x40); 		// point to the left margin of the 2nd line 
        putLCD(' ');
        
        // and draw just the tip appearing from the right
        setLCDC(0x40+15);		// point to the right margin of the 2nd line
        putLCD( 0);
        DELAY();
        
    } // repeat the main loop
} // main
